#!/bin/bash

##
## File: test_filedma.sh
## Author: glazzari
## Description: runs file-based regression tests for GMT, Calo and Passthrough processors.
## Style: Google Shell Guide (https://google.github.io/styleguide/shellguide.html).
## Gold standard for regression: scdaq @ 67ddf197 from January 9th, 2024.
##

readonly DATA_DIR=test/data
readonly CONFIG_DIR=test/config
readonly HASH_DIR=${DATA_DIR}/.sha256
readonly BUILD_DIR=build
readonly SCDAQ_TARGET=scdaq
readonly SCDAQ_PATH=${BUILD_DIR}/${SCDAQ_TARGET}
readonly MAX_RUNTIME_SECS=20
readonly MAX_HASHABLE_BYTES=104857600
readonly REF_COMMIT_TAG=67ddf197 # After the CMake migration

function build_scdaq() {
  mkdir -p ${BUILD_DIR}
  cd ${BUILD_DIR} && cmake .. && make ${SCDAQ_TARGET} && cd .. && echo "${FUNCNAME}: SUCCESS"
}

function cleanup_test() {
  cd ${DATA_DIR} && rm -rf *.dat *.journal in_progress run000* && cd - > /dev/null && echo "${FUNCNAME}: SUCCESS"
}

function setup_test() {
  local modify_config=${MODIFY_CONFIG}

  cleanup_test

  if [[ ${modify_config} = true ]]; then
      echo "${FUNCNAME}: using modified configuration with 5 orbits per packet"
      # Next line needed because older GMT files were taken with 5 orbits per packet
      sed -i 's/nOrbitsPerPacket: *1/nOrbitsPerPacket:5/' ${CONFIG_DIR}/scdaq-gmt.conf
      sed -i 's/nOrbitsPerPacket: *1/nOrbitsPerPacket:5/' ${CONFIG_DIR}/filedma-gmt.json5
      cd "${DATA_DIR}"
      tar -xzf calo_testfile_mod.tar.gz
      tar -xzf gmt_testfile_mod.tar.gz
  else
      cd "${DATA_DIR}"
      tar -xzf calo_testfile.tar.gz
      tar -xzf gmt_testfile.tar.gz
  fi
  cd - > /dev/null
  echo "${FUNCNAME}: SUCCESS"
}

function hash_first_n_bytes() {
  local num_bytes=$1
  local file_path=$2
  echo "$(head -c ${num_bytes} ${file_path} | sha256sum | cut -f1 -d' ' | cat)"
}

function check_hashes() {
  local num_bytes=$1
  local output_file=$2
  local reference_hash_file=$3

  output_hash=$(hash_first_n_bytes ${num_bytes} ${output_file})
  echo "${FUNCNAME}: computed hash is '${output_hash}'"

  if [ "$(cat ${reference_hash_file})" == "${output_hash}" ]; then
    echo "${FUNCNAME}: hashes match - SUCCESS"
    return 0
  else
    echo "${FUNCNAME}: hashes do not match!"
    echo "-- before changes: $(cat ${reference_hash_file})"
    echo "-- after changes:  ${output_hash}"
    return 1
  fi
}

function run_filedma_test() {
  local test_id=$1
  declare -i timeout_secs=$2
  declare -i num_streams=$3
  local config_file=$4

  local expects_output=${EXPECTS_OUTPUT}
  declare -i bytes_to_hash=${MAX_HASHABLE_BYTES}
  local ref_hash_file=${HASH_DIR}/${test_id}_${bytes_to_hash}_${REF_COMMIT_TAG}

  readonly test_id timeout_secs num_streams config_file expects_output ref_hash_file bytes_to_hash
  local output_dir output_prefix output_file ret_status

  setup_test

  output_dir=$(awk -F":" '/^output_filename_base:/{printf $2}' $config_file)
  output_prefix=$(awk -F":" '/^output_filename_prefix:/{printf $2}' $config_file)

  local test_failed=0
  echo "${FUNCNAME}: starting SCDAQ test with ${test_id} P5 capture file"
  echo "${FUNCNAME}: timeout ${timeout_secs}s ./${SCDAQ_PATH} --config ${config_file} --nstreams ${num_streams}"
  timeout ${timeout_secs}s ./"${SCDAQ_PATH}" --config ${config_file} --nstreams ${num_streams} | uniq -uc

  ret_status="${PIPESTATUS[0]}"

  # We expect a TIMEOUT failure because `scdaq` is not normally supposed to return.
  if [[ "${ret_status}" -ne 124 ]]; then
    echo "${FUNCNAME}: SCDAQ test for ${test_id} FAILED with exit code ${ret_status}!"
    test_failed=1
  # Check if output file was correctly generated
  elif [[ ${expects_output} = true ]]; then
    # Find file matching name pattern under the output directory, and excluding input data files
    # In case of multiple files, sorts alphanumerically so the lower-numbered one comes first
    local output_files=$(find ${output_dir} -iname "*${output_prefix}*.dat" ! -iname "*testfile.dat" | sort -rV)
    if [ -z "${output_files}" ]; then
      echo "${FUNCNAME}: SCDAQ test for ${test_id} FAILED: output file not found!"
    else
      echo "${FUNCNAME}: output file(s) found: ${output_files}."
      # More than one file may have been found; check by counting lines.
      local output_file
      if [[ $(echo "${output_files}" | wc -l) -gt 1 ]]; then
        # Since the list has been sorted, take the first (and "earliest") one with `head`
        output_file=$(echo "${output_files}" | head -n 1)
      else
        # Single matched file
        output_file="${output_files}"
      fi
      echo "${FUNCNAME}: taking file for comparison: ${output_file}."
      # Hash-verify contents
      if check_hashes ${bytes_to_hash} ${output_file} ${ref_hash_file} ; then
        echo "${FUNCNAME}: SCDAQ test PASSED for ${test_id}"
      else
        echo "${FUNCNAME}: SCDAQ test for ${test_id} FAILED"
        test_failed=1
      fi
    fi
  fi
  return ${test_failed}
}

function run_all_tests () {
  local test_description=$1
  local expects_output=${EXPECTS_OUTPUT}
  declare -i timeout_secs=${MAX_RUNTIME_SECS}
  readonly test_description timeout_secs expects_output
  local gmt_conf_failed gmt_json5_failed calo_conf_failed calo_json5_failed
  local passthrough_conf_failed passthrough_json5_failed multistream_conf_failed

  echo "${FUNCNAME}: running ${test_description}"

  run_filedma_test "GMT" ${timeout_secs} 0 "${CONFIG_DIR}/scdaq-gmt.conf"
  gmt_conf_failed=$?
  run_filedma_test "CALO" ${timeout_secs} 0 "${CONFIG_DIR}/scdaq-calo.conf"
  calo_conf_failed=$?
  run_filedma_test "PASS0_GMT" ${timeout_secs} 0 "${CONFIG_DIR}/scdaq-passthrough.conf"
  passthrough_conf_failed=$?

  run_filedma_test "GMT" ${timeout_secs} 0 "${CONFIG_DIR}/filedma-gmt.json5"
  gmt_json5_failed=$?
  run_filedma_test "CALO" ${timeout_secs} 0 "${CONFIG_DIR}/filedma-calo.json5"
  calo_json5_failed=$?
  run_filedma_test "PASS0_GMT" ${timeout_secs} 0 "${CONFIG_DIR}/filedma-passthrough.json5"
  passthrough_json5_failed=$?

  if [[ ${expects_output} = false ]]; then
    # We evaluate multi-stream operation with two parallel pipelines
    run_filedma_test "MULTI_PASS" ${timeout_secs} 2 "${CONFIG_DIR}/scdaq-passthrough.conf"
    multistream_conf_failed=$?
    run_filedma_test "MULTI_PASS" ${timeout_secs} 2 "${CONFIG_DIR}/filedma-passthrough.json5"
    multistream_json5_failed=$?
  fi

  local failed_any=0
  echo "${FUNCNAME}: error report for ${test_description} (0 = SUCCESS)"
  echo "                 conf | json5"
  echo "-- GMT         :  ${gmt_conf_failed}   |  ${gmt_json5_failed}"
  echo "-- Calo        :  ${calo_conf_failed}   |  ${calo_json5_failed}"
  echo "-- Passthrough :  ${passthrough_conf_failed}   |  ${passthrough_json5_failed}"
  if [[ ${expects_output} = false ]]; then
    echo "-- Multi-stream:  ${multistream_conf_failed}   |  ${multistream_json5_failed}"
    failed_any=${multistream_conf_failed}+${multistream_json5_failed}
  fi

  failed_any=${failed_any}+${gmt_json5_failed}+${gmt_conf_failed}+${calo_json5_failed}+${calo_conf_failed}+${passthrough_json5_failed}+${passthrough_conf_failed}
  if [[ $failed_any -ne 0 ]]; then
    echo "${FUNCNAME}: ${test_description} ===> FAILED!"
    exit 1
  else
    echo "${FUNCNAME}: ${test_description} ===> SUCCESS"
  fi
}

function main {
  echo "${FUNCNAME}: building scdaq locally"

  build_scdaq

  # Test 1: time-bound operation, checks for crashes
  MODIFY_CONFIG=false EXPECTS_OUTPUT=false run_all_tests "Test 1 (operation) - ${MAX_RUNTIME_SECS}-second runs, no output files"

  # Test 2: validates file output for basic runs
  MODIFY_CONFIG=true EXPECTS_OUTPUT=true run_all_tests "Test 2 (regression) - ${MAX_RUNTIME_SECS}-second runs, check output files"

  echo "${FUNCNAME}: all tests finished"
}

main "$@"