#ifndef CONFIG_H
#define CONFIG_H

#include <jsoncpp/json/json.h>

#include <boost/lexical_cast.hpp>
#include <cstdint>
#include <map>
#include <stdexcept>
#include <string>

#include "log.h"

class Config {
  class ConfigMap {
   public:
    using ValueType = Json::Value;
    using ConstIteratorType = Json::Value::const_iterator;

   private:
    ValueType data_;

   public:
    void Insert(const std::string &key, ValueType val) { data_[key] = std::move(val); }

    void Import(const ValueType &values) { data_ = values; }

    void Print() const {
      for (ConstIteratorType it = data_.begin(); it != data_.end(); it++) {
        auto k = it.key().asString();
        LOG(INFO) << "key: " << k << " | value: " << data_[k];
      }
    }

    std::string at(const std::string &key) const {
      try {
        auto &val = data_[key];
        return val.asString();
      } catch (std::exception &e) {
        LOG(ERROR) << "Error fetching value for key '" << key << "'";
        LOG(ERROR) << e.what();
        throw;
      }
    }
  };

 public:
  enum class InputType { WZDMA, FILEDMA, TCPIP, MICRONDMA, FILE };
  enum class ProcessorType { PASS_THROUGH, GMT, CALO, BRIL };
  enum class ConfigFormat { CONF, JSON5 };

  explicit Config(const std::string &filename);

  template <ConfigFormat>
  int Parse(std::ifstream &input);

  void print() const;

  InputType getInput() const {
    const std::string &input = dict_.at("input");
    if (input == "wzdma") {
      return InputType::WZDMA;
    }
    if (input == "filedma") {
      return InputType::FILEDMA;
    }
    if (input == "tcpip") {
      return InputType::TCPIP;
    }
    if (input == "micronDMA") {
      return InputType::MICRONDMA;
    }
    if (input == "file") {
      return InputType::FILE;
    }
    throw std::invalid_argument("Configuration error: Wrong input type '" + input + "'");
  }

  bool getDev_TCPAutoReconnectOnFailure() const {
    return (dict_.at("dev_TCPAutoReconnectOnFailure") == "true");
  }

  LOG_LEVEL getLogMinSeverity() const {
    const std::string &input = dict_.at("log_min_severity");
    if (input == "TRACE") {
      return TRACE;
    }
    if (input == "DEBUG") {
      return DEBUG;
    }
    if (input == "INFO") {
      return INFO;
    }
    if (input == "WARNING") {
      return WARNING;
    }
    if (input == "ERROR") {
      return ERROR;
    }
    if (input == "FATAL") {
      return FATAL;
    }
    throw std::invalid_argument("Configuration error: Unknown log_min_severity '" + input + "'");
  }

  std::string getDmaDevice() const { return dict_.at("dma_dev"); }

  uint64_t getTcpDestPort() const { return boost::lexical_cast<uint32_t>(dict_.at("tcpDestPort")); }

  uint64_t getDmaPacketBufferSize() const {
    return boost::lexical_cast<uint64_t>(dict_.at("dma_packet_buffer_size"));
  }

  uint32_t getNumberOfDmaPacketBuffers() const {
    return boost::lexical_cast<uint32_t>(dict_.at("dma_number_of_packet_buffers"));
  }

  uint32_t getPacketsPerReport() const {
    return boost::lexical_cast<uint32_t>(dict_.at("packets_per_report"));
  }

  std::string getInputFile() const { return dict_.at("input_file"); }

  ProcessorType getProcessorType() const {
    const std::string &input = dict_.at("processor_type");
    if (input == "PASS_THROUGH") {
      return ProcessorType::PASS_THROUGH;
    }
    if (input == "GMT") {
      return ProcessorType::GMT;
    }
    if (input == "CALO") {
      return ProcessorType::CALO;
    }
    if (input == "BRIL") {
      return ProcessorType::BRIL;
    }
    throw std::invalid_argument("Configuration error: Wrong processor type '" + input + "'");
  }

  std::string getOutputFilenamePrefix() const { return dict_.at("output_filename_prefix"); }

  std::string getOutputFilenameBase() const { return dict_.at("output_filename_base"); }

  uint64_t getOutputMaxFileSize() const {
    return boost::lexical_cast<uint64_t>(dict_.at("max_file_size"));
  }

  bool getOutputForceWrite() const { return (dict_.at("output_force_write") == "yes"); }

  uint32_t getNOrbitsPerPacket() const {
    return boost::lexical_cast<uint32_t>(dict_.at("nOrbitsPerPacket"));
  }

  uint32_t getNumThreads() const { return boost::lexical_cast<uint32_t>(dict_.at("threads")); }

  short getPortNumber() const { return boost::lexical_cast<short>(dict_.at("port")); }

  bool getEnableStreamProcessor() const { return (dict_.at("enable_stream_processor") == "yes"); }

  bool getDoZS() const { return (dict_.at("doZS") == "yes"); }

  uint32_t getSourceID() const;

  bool getCMSSWHeaders() const { return (dict_.at("cmsswHeaders") == "yes"); }

  uint32_t getNOrbitsPerFile() const {  // If > 0, the fixed N orbits per file option is performed
    return boost::lexical_cast<uint32_t>(dict_.at("nOrbitsPerFile"));
  }

  std::string getSconeHost() const { return dict_.at("scone_host"); }

  std::string getSconePort() const { return dict_.at("scone_port"); }

  std::string getSconeBoard() const { return dict_.at("scone_board"); }

 private:
  ConfigMap dict_;
};
#endif
