#include "pipeline.h"

#include "bril_processor.h"
#include "calo_orbit_processor.h"
#include "muon_orbit_processor.h"
#include "orbit_processor.h"
#include "processor.h"

template <typename T>
std::shared_ptr<OrbitProcessor> MakeA(const Config &conf, ctrl &control) {
  return std::make_shared<T>(conf.getNOrbitsPerPacket(), conf.getDoZS(), conf.getCMSSWHeaders(),
                             static_cast<uint16_t>(conf.getSourceID()), control);
}

std::shared_ptr<Processor> MakeProcessor(const Config &conf, ctrl &control) {
  switch (conf.getProcessorType()) {
    case Config::ProcessorType::GMT:
      return MakeA<MuonOrbitProcessor>(conf, control);
    case Config::ProcessorType::CALO:
      return MakeA<CaloOrbitProcessor>(conf, control);
    case Config::ProcessorType::BRIL:
      return std::make_shared<BrilProcessor>(control);
    case Config::ProcessorType::PASS_THROUGH:
      return std::make_shared<Processor>();
    default:
      throw std::invalid_argument("Error: Unrecognized processor type.");
  }
}

Pipeline::Pipeline(Config &conf, uint maxTokensPerThread)
    : fMaxTokensPerThread(maxTokensPerThread), fConf(conf) {
  tools::log::set_filter_min_severity(fConf.getLogMinSeverity());
  fConf.print();
  const Config::InputType &input = fConf.getInput();
  const size_t &nbPacketBuffers = fConf.getNumberOfDmaPacketBuffers();
  const size_t &packetBufferSize = fConf.getDmaPacketBufferSize();

  if (input == Config::InputType::FILEDMA) {
    // Create FILE DMA reader
    fInputFilter = std::make_shared<FileDmaInputFilter>(fConf.getInputFile(), packetBufferSize,
                                                        nbPacketBuffers, control);

  } else if (input == Config::InputType::WZDMA) {
    // Create WZ DMA reader
    fInputFilter =
        std::make_shared<WZDmaInputFilter>(packetBufferSize, nbPacketBuffers, control, fConf);

  } else if (input == Config::InputType::MICRONDMA) {
    // create MicronDmaInputFilter reader
    fInputFilter =
        std::make_shared<MicronDmaInputFilter>(packetBufferSize, nbPacketBuffers, control, fConf);

  } else if (input == Config::InputType::TCPIP) {
    // create TcpInputFilter reader
    if (fConf.getDev_TCPAutoReconnectOnFailure()) {
      fInputFilter =
          std::make_shared<TcpInputFilter>(fConf.getTcpDestPort(), packetBufferSize,
                                           nbPacketBuffers, fConf.getNOrbitsPerPacket(), control);

    } else {
      throw std::invalid_argument(
          "Configuration error: enable developmentMode to use TCP input filter");
    }
  } else {
    throw std::invalid_argument("Configuration error: Unknown input type was specified");
  }

  // Add input reader to a pipeline
  fPipeline.add_filter(*fInputFilter);

  // Create reformatter and add it to the pipeline
  if (fConf.getEnableStreamProcessor()) {
    fStreamProcessor = MakeProcessor(fConf, control);
    fPipeline.add_filter(*fStreamProcessor);
  }

  // Create file-writing stage and add it to the pipeline
  if (fConf.getNOrbitsPerFile()) {
    fOutputStream = std::make_shared<OutputByOrbitStream>(control, fConf);
  } else {
    fOutputStream = std::make_shared<OutputBySizeStream>(
        fConf.getOutputFilenameBase(), fConf.getOutputFilenamePrefix(), Pipeline::control,
        fConf.getProcessorType());
  }
  fPipeline.add_filter(*fOutputStream);

  //    tbb::tick_count mainStartTime = tbb::tick_count::now();
  control.running = false;
  control.run_number = 0;
  control.max_file_size = fConf.getOutputMaxFileSize();  // in Bytes
  control.packets_per_report = static_cast<int>(fConf.getPacketsPerReport());
  control.output_force_write = fConf.getOutputForceWrite();
  control.n_orbits_per_packet = static_cast<int>(fConf.getNOrbitsPerPacket());

  // Firmware needs at least 1MB buffer for DMA
  if (fConf.getDmaPacketBufferSize() < 1024 * 1024) {
    LOG(ERROR) << "dma_packet_buffer_size must be at least 1048576 bytes (1MB), but "
               << fConf.getDmaPacketBufferSize()
               << " bytes was given. Check the configuration file.";
    throw std::invalid_argument("Configuration error: DMA packet buffer size too small. ");
  }
}

MultiPipeline::MultiPipeline(std::vector<Config> &configs) {
  for (auto &config : configs) {
    try {
      fPipelines.emplace_back(config);
      LOG(DEBUG) << "-Configuration loaded";
      LOG(DEBUG) << "-Pipeline set up";
    } catch (std::invalid_argument &e) {
      LOG(FATAL) << "-Configuration invalid! Error text is \"" << e.what() << "\" Bailing out.";
      throw;
    } catch (std::exception &e) {
      LOG(ERROR) << "-Error occurred. error text is: \"" << e.what() << "\"";
      throw;
    }
  }

  {  // RAII
    fServer = std::unique_ptr<server>(
        new server(io_service, fPipelines[0].fConf.getPortNumber(), Pipeline::control));
    fServerThread = boost::thread(boost::bind(&boost::asio::io_service::run, &io_service));
  }
}

int MultiPipeline::Run() {
  int retval = 0;
  assert(!fPipelines.empty());
  try {
    if (!RunPipelines(static_cast<unsigned int>(fPipelines.size()))) {
      retval = 1;  // Will terminate with error code set
    }
  } catch (std::exception &e) {
    LOG(ERROR) << "Error in pipelines. Error text is: \"" << e.what() << "\"";
    retval = 1;                             // Will terminate with error code set
    Config &conf = fPipelines.at(0).fConf;  // Use info from first file for SCONE reset
    LOG(INFO) << "Resetting board at http://" << conf.getSconeHost() << ":" << conf.getSconePort()
              << "/" << conf.getSconeBoard() << " before exiting.. ";
    int res = tools::reset_board(conf.getSconeHost(), conf.getSconePort(), conf.getSconeBoard());
    if (res != 0) {
      LOG(ERROR) << "Reset failed! Error code: " << res;
    } else {
      LOG(INFO) << "Successfully reset the board.";
    }
  }

  LOG(DEBUG) << "Terminating the internal TCP server.";
  io_service.stop();
  if (!fServerThread.try_join_for(boost::chrono::milliseconds(2000))) {
    LOG(ERROR) << "Failed to terminate the internal TCP server, the program "
                  "may crash on exit.";
  } else {
    LOG(DEBUG) << "Internal TCP server is terminated.";
  }

  return retval;
}

int MultiPipeline::RunPipelines(uint numStreams) {
  tbb::task_group group;
  // Run pipelines and measure real time
  tbb::tick_count t0 = tbb::tick_count::now();
  for (uint i = 0; i < numStreams; ++i) {
    auto &p = this->fPipelines.at(i);
    group.run([&] { return p.fPipeline.run(p.fConf.getNumThreads() * p.fMaxTokensPerThread); });
  }
  group.wait();
  tbb::tick_count t1 = tbb::tick_count::now();

  if (fVerbose) {
    LOG(INFO) << "time = " << (t1 - t0).seconds();
  }
  return 1;
}